from abc import abstractmethod


class Effect:
    @abstractmethod
    def apply(self, data):
        pass

    def __rshift__(self, other):
        return ChainedEffect(self, other)


class ChainedEffect(Effect):
    def __init__(self, effect1, effect2):
        self.effect1 = effect1
        self.effect2 = effect2

    def apply(self, data):
        return self.effect2.apply(self.effect1.apply(data))



# SPDX-FileCopyrightText: 2022-present dulong-lab <dulong-lab@outlook.com>
#
# SPDX-License-Identifier: MIT
from .const import *
from .oscillator import *
from .instrument import *
from .effect import *
from .let import *
from .note import *
from .tempo import *
from .audio_compiler import *
from .tools import *
from .score import *



from typing import Iterable


class Tempo:
    def __init__(self, number, referent=1/4, text=None):
        self.number = number
        self.referent = referent
        self.text = text
        self.__seconds_per_whole_note = 60 / (number * referent)

    def convert(self, dur):
        return self.__seconds_per_whole_note * dur

    @classmethod
    def create(cls, args):
        if isinstance(args, (int, float)):
            return Tempo(args)
        elif isinstance(args, Iterable):
            return Tempo(*args)
        else:
            return args


class NamedTempo(Tempo):
    def __call__(self, number, referent=None):
        return NamedTempo(number, referent or self.referent, self.text)


Allegro = NamedTempo(120, 1/4, "Allegro")

__all__ = ['Tempo', 'Allegro']



from functools import cached_property
from inspect import signature
from typing import Iterable

import matplotlib.pyplot as plt
import numpy as np

from .const import SAMPLING_RATE


class Oscillator:
    @cached_property
    def map(self):
        argc = len(signature(self.take).parameters)  # type: ignore
        if argc == 1:
            return self.__map1
        elif argc == 2:
            return self.__map2
        elif argc == 3:
            return self.__map3
        else:
            raise Exception("take 参数数量错误")

    @cached_property
    def __vectorized_take(self):
        return np.vectorize(self.take, cache=True)  # type: ignore

    def __map1(self, p, _, __):
        return self.__vectorized_take(p)

    def __map2(self, p, t, _):
        return self.__vectorized_take(p, t)

    def __map3(self, p, t, d):
        return self.__vectorized_take(p, t, d)

    def plot(self, n=1, dur=1.0):
        s = int(dur * SAMPLING_RATE)
        p = np.tile(np.linspace(0, 1, s // n, endpoint=False), n + 1)[:s]
        t = np.linspace(0, dur, s, endpoint=False)
        d = np.full(s, t[-1])
        data = np.array(self.map(p, t, d))

        if len(data.shape) == 1:
            plt.plot(data)
        else:
            for row in data:
                plt.plot(row)

    def __combine(self, op, other):
        if isinstance(other, Oscillator):
            return OscillatorOpOscillator(self, op, other)
        else:
            return OscillatorOpValue(self, op, other)

    def __add__(self, other):
        return self.__combine(np.add, other)

    def __sub__(self, other):
        return self.__combine(np.subtract, other)

    def __mul__(self, other):
        return self.__combine(np.multiply, other)

    def __pow__(self, other):
        return self.__combine(np.power, other)

    def __truediv__(self, other):
        return self.__combine(np.true_divide, other)

    def __floordiv__(self, other):
        return self.__combine(np.floor_divide, other)

    def __mod__(self, other):
        return self.__combine(np.mod, other)

    def __pos__(self):
        return self

    def __neg__(self):
        return OpOscillator(np.negative, self)

    def __abs__(self):
        return OpOscillator(np.abs, self)


class CombinedOscillator(Oscillator):
    def take(self, p, t, d):
        v = self.map([p], [t], [d]).T[0]
        return tuple(v) if isinstance(v, Iterable) else v


class OpOscillator(CombinedOscillator):
    def __init__(self, op, a):
        self.op = op
        self.a = a

    def map(self, p, t, d):
        return self.op(self.a.map(p, t, d))


class OscillatorOpOscillator(CombinedOscillator):
    def __init__(self, a, op, b):
        self.a = a
        self.op = op
        self.b = b

    def map(self, p, t, d):
        return self.op(self.a.map(p, t, d), self.b.map(p, t, d))


class OscillatorOpValue(CombinedOscillator):
    def __init__(self, a, op, b):
        self.a = a
        self.op = op
        self.b = b

    def map(self, p, t, d):
        return self.op(self.a.map(p, t, d), self.b)


__all__ = ['Oscillator']



from copy import copy
from inspect import isclass

import IPython.display as ipd
import numpy as np

from .let import Let
from .note import NamedNote, Note, R
from .tempo import Tempo


class AudioCompiler:
    def __init(self, instrument, tempo):
        self.__instrument = instrument
        self.__tempo = Tempo.create(tempo)
        self.__saved_instrument = instrument
        self.__saved_tempo = tempo
        self.__p8 = None
        self.__dur = None
        self.__nchannel = None
        self.__rest_idx = []
        self.__let_idx = []

    def compile(self, instrument, tempo, symbols):
        self.__init(instrument, tempo)
        data = [self.__compile_symbol(i, it) for i, it in enumerate(symbols)]

        if self.__nchannel == None:
            self.__nchannel = self.__guess_nchannel()

        if self.__nchannel != 1:
            for i in self.__rest_idx:
                data[i] = np.zeros([self.__nchannel, len(data[i])])
            for i in self.__let_idx:
                data[i] = np.empty([self.__nchannel, 0])

        if len(data) == 0:
            return np.empty(0 if self.__nchannel == 1 else [self.__nchannel, 0])

        return np.concatenate(data, axis=0 if self.__nchannel == 1 else 1)

    def audio(self, instrument, tempo, symbols, autoplay=False, normalize=False):
        data = self.compile(instrument, tempo, symbols)
        return ipd.Audio(
            data if len(data) != 0 else [0],
            rate=instrument.sr,
            autoplay=autoplay,
            normalize=normalize
        )

    def __guess_nchannel(self):
        if self.__instrument.wav == None:
            # Score 创建的 Instrument 默认 wav 为 None
            return 1
        else:
            return len(self.__instrument.wav.map([0], [0], [0]))

    def __compile_symbol(self, i, symbol):
        if isclass(symbol):
            if issubclass(symbol, NamedNote):
                note = symbol().complete(self.__p8, self.__dur)  # type: ignore
                return self.__compile_named_note(note)
            elif issubclass(symbol, R):
                return self.__compile_rest(i, symbol().complete(self.__dur))
            elif issubclass(symbol, Note):
                return self.__compile_note(symbol().complete(self.__dur))
            else:
                raise Exception(f"不明符号 {symbol}")
        else:
            if isinstance(symbol, NamedNote):
                return self.__compile_named_note(symbol.complete(self.__p8, self.__dur))
            elif isinstance(symbol, R):
                return self.__compile_rest(i, symbol.complete(self.__dur))
            elif isinstance(symbol, Note):
                return self.__compile_note(symbol.complete(self.__dur))
            elif isinstance(symbol, Let):
                self.__let_idx.append(i)
                self.__let_context(symbol)
                return np.empty(0)
            else:
                raise Exception(f"不明符号 {symbol}")

    def __compile_named_note(self, note):
        self.__before_produce(note)
        data = self.__instrument.produce(
            note.freq, self.__tempo.convert(note.dur))
        self.__after_produce(note)

        self.__p8 = note.p8
        self.__dur = note.dur

        if self.__nchannel == None:
            self.__nchannel = 1 if len(data.shape) == 1 else data.shape[0]

        return data

    def __compile_rest(self, i, note):
        self.__before_produce(note)
        data = np.zeros(int(self.__tempo.convert(
            note.dur) * self.__instrument.sr))
        self.__after_produce(note)

        self.__dur = note.dur
        self.__rest_idx.append(i)

        return data

    def __compile_note(self, note):
        self.__before_produce(note)
        data = self.__instrument.produce(
            note.freq, self.__tempo.convert(note.dur))
        self.__after_produce(note)

        self.__dur = note.dur
        if self.__nchannel == None:
            self.__nchannel = 1 if len(data.shape) == 1 else data.shape[0]

        return data

    def __before_produce(self, note):
        if note.after_let:
            self.__let_context(note.after_let)
        if note.that_let:
            self.__save_context()
            self.__let_context(note.that_let)

    def __after_produce(self, note):
        if note.that_let:
            self.__load_context()
        if note.then_let:
            self.__let_context(note.then_let)

    def __save_context(self):
        self.__saved_tempo = self.__tempo
        self.__instrument, self.__saved_instrument =\
            copy(self.__instrument), self.__instrument

    def __let_context(self, let):
        if let.wav != None:
            self.__instrument.wav = let.wav
        if let.amp != None:
            self.__instrument.amp = let.amp
        if let.freq != None:
            self.__instrument.freq = let.freq
        if let.fx != None:
            self.__instrument.fx = let.fx
        if let.tempo != None:
            self.__tempo = Tempo.create(let.tempo)
        if let.p8 != None:
            self.__p8 = let.p8
        if let.dur != None:
            self.__dur = let.dur

    def __load_context(self):
        self.__tempo = self.__saved_tempo
        self.__instrument = self.__saved_instrument


__all__ = ['AudioCompiler']



from copy import copy
from functools import cached_property

from .let import Let


class NoteMeta(type):
    def __truediv__(cls, value):
        return cls().__truediv__(value)


class Note(metaclass=NoteMeta):
    def __init__(self):
        self.dur = None
        self.after_let = None
        self.that_let = None
        self.then_let = None
        self.after = self.__after
        self.that = self.__that
        self.then = self.__then

    def __truediv__(self, value):
        obj = copy(self)
        obj.dur = 1.0 / value
        return obj

    def __mul__(self, value):
        obj = copy(self)
        obj.dur *= value
        return obj

    @classmethod
    def after(cls, wav=None, amp=None, freq=None, fx=None):
        return cls().__after(wav, amp, freq, fx)

    @classmethod
    def that(cls, wav=None, amp=None, freq=None, fx=None):
        return cls().__that(wav, amp, freq, fx)

    @classmethod
    def then(cls, wav=None, amp=None, freq=None, fx=None):
        return cls().__then(wav, amp, freq, fx)

    def __after(self, wav=None, amp=None, freq=None, fx=None):
        obj = copy(self)
        obj.after_let = Let(wav=wav, amp=amp, freq=freq, fx=fx)
        return obj

    def __that(self, wav=None, amp=None, freq=None, fx=None):
        obj = copy(self)
        obj.that_let = Let(wav=wav, amp=amp, freq=freq, fx=fx)
        return obj

    def __then(self, wav=None, amp=None, freq=None, fx=None):
        obj = copy(self)
        obj.then_let = Let(wav=wav, amp=amp, freq=freq, fx=fx)
        return obj

    def complete(self, dur):
        if self.dur != None:
            return self

        obj = copy(self)
        obj.dur = dur
        return obj


class N(Note):
    def __init__(self, freq):
        super().__init__()
        self.freq = freq


class R(Note):
    pass


class NamedNoteMeta(NoteMeta):
    def __pos__(cls):
        return cls().__pos__()

    def __neg__(cls):
        return cls().__neg__()


class NamedNote(Note, metaclass=NamedNoteMeta):
    def __init__(self, std_freq, p8):
        super().__init__()
        self.std_freq = std_freq
        self.p8 = p8
        self.accidental = 0

    def __pos__(self):
        # TODO: 错误检查
        obj = copy(self)
        obj.accidental += 1
        return obj

    def __neg__(self):
        # TODO: 错误检查
        obj = copy(self)
        obj.accidental -= 1
        return obj

    def complete(self, p8, dur):
        if self.p8 != None and self.dur != None:
            return self

        obj = copy(self)
        if obj.p8 == None:
            obj.p8 = p8
        if obj.dur == None:
            obj.dur = dur
        return obj

    @cached_property
    def freq(self):
        return self.std_freq * 2**(self.p8-4 + self.accidental/12)


class C(NamedNote):
    def __init__(self, p8=None):
        super().__init__(440.0 * 2**(-9/12), p8)


class D(NamedNote):
    def __init__(self, p8=None):
        super().__init__(440.0 * 2**(-7/12), p8)


class E(NamedNote):
    def __init__(self, p8=None):
        super().__init__(440.0 * 2**(-5/12), p8)


class F(NamedNote):
    def __init__(self, p8=None):
        super().__init__(440.0 * 2**(-4/12), p8)


class G(NamedNote):
    def __init__(self, p8=None):
        super().__init__(440.0 * 2**(-2/12), p8)


class A(NamedNote):
    def __init__(self, p8=None):
        super().__init__(440.0, p8)


class B(NamedNote):
    def __init__(self, p8=None):
        super().__init__(440.0 * 2**(2/12), p8)


__all__ = ['Note', 'N', 'R', 'C', 'D', 'E', 'F', 'G', 'A', 'B']



import IPython.display as ipd
import numpy as np

from .audio_compiler import AudioCompiler
from .const import SAMPLING_RATE
from .tempo import Tempo
from .instrument import Instrument
from .tools import normalize


class Score:
    class Mixer:
        def mix(self, tracks):
            return normalize(sum(tracks.values()))

    def __init__(self, tempo, parts, title=None, composer=None, mixer=Mixer(), sr=SAMPLING_RATE) -> None:
        self.title = title
        self.composer = composer
        self.mixer = mixer
        self.sr = sr
        self.parts = parts
        self.tempo = Tempo.create(tempo)
        self.instruments = {part: Instrument(None, sr=sr) for part in parts}
        self.symbols = {part: [] for part in parts}

    def audio(self, parts=None, autoplay=False, normalize=False):
        data = self.build(parts)
        return ipd.Audio(
            data if len(data) != 0 else [0],  # type: ignore
            rate=self.sr,
            autoplay=autoplay,
            normalize=normalize
        )

    def build(self, parts=None):
        return self.mixer.mix(self.__pad(self.__normalize(self.compile(parts))))

    def compile(self, parts=None):
        if parts == None:
            parts = self.parts
        elif (not set(parts).issubset(self.parts)):
            raise Exception("指定了无效的声部")

        return {
            part: AudioCompiler().compile(
                self.instruments[part],
                self.tempo,
                self.symbols[part] if part in parts else []
            )
            for part in self.parts
        }

    def __pad(self, tracks):
        width = max(map(lambda a: a.shape[-1], tracks.values()))
        return {
            part: self.__pad_to_width(data, width)
            for part, data in tracks.items()
        }

    def __pad_to_width(self, data, width):
        d = len(data.shape)
        n = width - data.shape[0]
        pad_width = np.array([(0, 0)]*(d-1) + [(0, n)])
        return np.pad(data, pad_width)

    def __normalize(self, tracks):
        # normalize 防止有时因浮点数经度问题导致出现不在 [-1, 1] 内的值
        # 同时让 mixer 获取到的数据更统一
        return {
            part: normalize(data)
            for part, data in tracks.items()
        }

    def __getitem__(self, name):
        return self.symbols[name]

    def __setitem__(self, name, value):
        self.symbols[name] = value


__all__ = ['Score']



import IPython.display as ipd
import matplotlib.pylab as plt
import numpy as np

from .const import SAMPLING_RATE
from .oscillator import Oscillator


class Instrument:
    def __init__(self, wav, amp=1, freq=1, fx=None, sr=SAMPLING_RATE):
        self.wav = wav
        self.amp = amp
        self.freq = freq
        self.fx = fx
        self.sr = sr

    def produce(self, freq, dur):
        n = int(dur * self.sr)
        p = np.linspace(0, 1, n, endpoint=False)
        t = np.linspace(0, dur, n, endpoint=False)
        d = np.full(n, t[-1])
        f = self.__carrier(self.freq, p, t, d, n) * freq
        a = self.__carrier(self.amp, p, t, d, n)

        data = self.wav.map(np.modf(t * f)[0], t, d)
        return (self.fx.apply(data) if self.fx else data) * a

    def plot(self, freq=1, dur=2):
        data = self.produce(freq, dur)

        if len(data.shape) == 1:
            plt.plot(data)
        else:
            for row in data:
                plt.plot(row)

    def audio(self, freq=440, dur=2, autoplay=False, normalize=False):
        data = self.produce(freq, dur)
        return ipd.Audio(
            data if len(data) != 0 else [0],
            rate=self.sr,
            autoplay=autoplay,
            normalize=normalize
        )

    def __carrier(self, obj, p, t, d, n):
        if isinstance(obj, Oscillator):
            return obj.map(p, t, d)
        else:
            return np.full(n, obj)


__all__ = ['Instrument']



class Let:
    def __init__(self, wav=None, amp=None, freq=None, fx=None, tempo=None, p8=None, dur=None) -> None:
        self.wav = wav
        self.amp = amp
        self.freq = freq
        self.fx = fx
        self.tempo = tempo
        self.p8 = p8
        self.dur = dur


let = Let

__all__ = ['let']
